@@ -9,32 +9,44 @@
 # stdlib
 import concurrent.futures
 import datetime
-from ftplib import error_reply, error_temp, error_perm, error_proto
 import getpass
 import logging
 import os
 import re
 import shutil
-from socket import timeout as socket_timeout
-from socket import create_connection
+import socket
 import sys
 import tempfile
 import traceback
 from contextlib import contextmanager
+from ftplib import error_perm, error_proto, error_reply, error_temp
 from math import ceil
+from socket import timeout as socket_timeout
 
 # 3rd party
 from paramiko.ssh_exception import SSHException
 
 # local modules
-from splitcopy.paramikoshell import SSHShell
 from splitcopy.ftp import FTP
 
 logger = logging.getLogger(__name__)
 
 
+def pad_string(text):
+    """pads a given string to the terminal width
+    :param text:
+    :type string:
+    :return padded_string:
+    :type string
+    """
+    term_width = shutil.get_terminal_size()[0]
+    padding = " " * (term_width - len(text))
+    padded_string = f"{text}{padding}"
+    return padded_string
+
+
 class SplitCopyShared:
-    """Class containing functions used by both SplitCopyGet
+    """class containing functions used by both SplitCopyGet
     and SplitCopyPut classes
     """
 
@@ -45,19 +57,20 @@ class SplitCopyShared:
         self.passwd = kwargs.get("passwd")
         self.ssh_key = kwargs.get("ssh_key")
         self.ssh_port = kwargs.get("ssh_port")
-        self.remote_dir = kwargs.get("remote_dir")
-        self.remote_file = kwargs.get("remote_file")
-        self.remote_path = kwargs.get("remote_path")
         self.local_dir = kwargs.get("local_dir")
         self.get_op = kwargs.get("get")
+        self.remote_dir = ""
+        self.remote_file = ""
         self.command_list = []
         self.rm_remote_tmp = False
         self.local_tmpdir = None
         self.remote_tmpdir = None
         self.sshshell = None
 
-    def connect(self, **ssh_kwargs):
-        """Function to open an ssh session to a remote host
+    def connect(self, ssh_lib, **ssh_kwargs):
+        """open an ssh session to a remote host
+        :param ssh_lib:
+        :type class:
         :param ssh_kwargs:
         :type dict:
         :return self.sshshell:
@@ -65,8 +78,11 @@ class SplitCopyShared:
         :returm ssh_kwargs:
         :type dict:
         """
+        logger.info("entering connect()")
         try:
-            self.sshshell = SSHShell(**ssh_kwargs)
+            self.sshshell = ssh_lib(**ssh_kwargs)
+            self.sshshell.socket_open()
+            self.sshshell.transport_open()
             if self.sshshell.main_thread_auth():
                 self.sshshell.channel_open()
                 self.sshshell.invoke_shell()
@@ -87,7 +103,7 @@ class SplitCopyShared:
         return self.sshshell, ssh_kwargs
 
     def which_proto(self, copy_proto):
-        """Function that determines which protocol will be used for the transfer.
+        """determines which protocol will be used for the transfer.
         If FTP is selected as protocol, verify that authentication works
         :param copy_proto:
         :type string:
@@ -96,6 +112,7 @@ class SplitCopyShared:
         :return passwd:
         :type string:
         """
+        logger.info("entering which_proto()")
         passwd = self.sshshell.kwargs["password"]
         result = None
         if copy_proto == "ftp" and self.ftp_port_check():
@@ -120,17 +137,18 @@ class SplitCopyShared:
         logger.info(f"copy_proto == {copy_proto}")
         return copy_proto, passwd
 
-    def ftp_port_check(self):
-        """Function that checks whether the ftp port is open
+    def ftp_port_check(self, socket_lib=socket):
+        """checks whether the ftp port is open
         :return result:
         :type bool:
         """
+        logger.info("entering ftp_port_check()")
         result = False
         print("attempting FTP authentication...")
         try:
-            with create_connection((self.host, 21), 10) as ftp_sock:
-                logger.info("ftp port is open")
-                result = True
+            socket_lib.create_connection((self.host, 21), 10)
+            logger.info("ftp port is open")
+            result = True
         except socket_timeout:
             print("ftp socket timed out, switching to scp for transfer")
         except ConnectionRefusedError:
@@ -138,13 +156,14 @@ class SplitCopyShared:
 
         return result
 
-    def ftp_login_check(self, passwd):
-        """Function that verifies ftp authentication on remote host
+    def ftp_login_check(self, passwd, ftp_lib=FTP):
+        """verifies ftp authentication on remote host
         :param passwd:
         :type string:
         :return result:
         :type bool:
         """
+        logger.info("entering ftp_login_check()")
         result = False
         kwargs = {
             "host": self.host,
@@ -152,12 +171,12 @@ class SplitCopyShared:
             "passwd": passwd,
             "timeout": 10,
         }
-        with FTP(**kwargs) as ftp:
+        with ftp_lib(**kwargs) as ftp:
             result = True
         return result
 
     def which_os(self):
-        """Function determines if host is JUNOS/EVO/*nix
+        """determines if host is JUNOS/EVO/*nix
         no support for remote Windows OS running OpenSSH
         :return junos:
         :type bool:
@@ -169,6 +188,7 @@ class SplitCopyShared:
         :type float:
         """
         logger.info("entering which_os()")
+        junos = False
         evo = False
         bsd_version = float()
         sshd_version = float()
@@ -177,7 +197,7 @@ class SplitCopyShared:
         if not result:
             err = "failed to determine remote host os, it must be *nix based"
             self.close(err_str=err)
-        host_os = stdout[0].split("\n")[0].rstrip()
+        host_os = stdout.split("\n")[1].rstrip()
         if host_os == "Linux" and self.evo_os():
             evo = True
         else:
@@ -189,7 +209,7 @@ class SplitCopyShared:
         return junos, evo, bsd_version, sshd_version
 
     def evo_os(self):
-        """Function that determines if host is running EVO
+        """determines if host is running EVO
         :return result:
         :type bool:
         """
@@ -198,7 +218,7 @@ class SplitCopyShared:
         return result
 
     def junos_os(self):
-        """Function that determines if host is running JUNOS
+        """determines if host is running JUNOS
         and if so which bsd and sshd versions are in use
         :return junos:
         :type bool:
@@ -228,7 +248,7 @@ class SplitCopyShared:
         return junos, bsd_version, sshd_version
 
     def which_bsd(self):
-        """Function that determines the BSD version of JUNOS
+        """determines the BSD version of JUNOS
         :return bsd_version:
         :type float:
         """
@@ -241,7 +261,7 @@ class SplitCopyShared:
         return bsd_version
 
     def which_sshd(self):
-        """Function that determines the OpenSSH daemon version
+        """determines the OpenSSH daemon version
         :return sshd_version:
         :type float
         """
@@ -255,7 +275,7 @@ class SplitCopyShared:
         return sshd_version
 
     def req_binaries(self, get_op=False, junos=False, evo=False):
-        """Function ensures required binaries exist on remote host
+        """ensures required binaries exist on remote host
         :param get_op:
         :type bool:
         :param junos:
@@ -278,14 +298,6 @@ class SplitCopyShared:
                     )
                 )
 
-    def second_elem(self, elem):
-        """Function used for key sort
-        :param elem:
-        :type list:
-        :return: the 2nd element of the list
-        """
-        return elem[1]
-
     def req_sha_binaries(self, sha_hash):
         """ensures required binaries for sha hash creation exist on remote host
         :param sha_hash:
@@ -315,7 +327,7 @@ class SplitCopyShared:
             bins = [("sha1sum", 1), ("sha1", 1), ("shasum", 1)]
             sha_bins.extend(bins)
 
-        sha_bins = sorted(set(sha_bins), reverse=True, key=self.second_elem)
+        sha_bins = sorted(set(sha_bins), reverse=True, key=lambda x: (x[1], x[0]))
         logger.info(sha_bins)
 
         for req_bin in sha_bins:
@@ -334,7 +346,7 @@ class SplitCopyShared:
         return sha_bin, sha_len
 
     def close(self, err_str=None, config_rollback=True, hard_close=False):
-        """Called when we want to exit the script
+        """called when we want to exit the script
         attempts to delete the remote temp directory and close the TCP session
         If hard_close == False, contextmanager will rm the local temp dir
         If not, we must delete it manually
@@ -347,13 +359,14 @@ class SplitCopyShared:
         :raises SystemExit: terminates the script gracefully
         :raises os._exit: terminates the script immediately (even asyncio loop)
         """
+        logger.info("entering close()")
         if err_str:
             print(err_str)
         if self.rm_remote_tmp:
             self.remote_cleanup()
         if config_rollback and self.command_list:
             self.limits_rollback()
-        print(f"\r{self.pad_string('closing device connection')}")
+        print(f"\r{pad_string('closing device connection')}")
         self.sshshell.close()
         if hard_close:
             try:
@@ -367,20 +380,8 @@ class SplitCopyShared:
         else:
             raise SystemExit(1)
 
-    def pad_string(self, text):
-        """Function that pads a given string to the terminal width
-        :param text:
-        :type string:
-        :return padded_string:
-        :type string
-        """
-        term_width = shutil.get_terminal_size()[0]
-        padding = " " * (term_width - len(text))
-        padded_string = f"{text}{padding}"
-        return padded_string
-
     def file_split_size(self, file_size, sshd_version, bsd_version, evo, copy_proto):
-        """Function determines the optimal chunk size. This depends on the python
+        """determines the optimal chunk size. This depends on the python
         version, cpu count, the protocol used and the FreeBSD/OpenSSH versions
         :returns split_size:
         :type int:
@@ -389,10 +390,11 @@ class SplitCopyShared:
         """
         logger.info("entering file_split_size()")
 
+        cpu_count = 1
         try:
             cpu_count = os.cpu_count()
         except NotImplementedError:
-            cpu_count = 1
+            pass
         max_workers = min(32, cpu_count * 5)
 
         # each uid can have max of 64 processes
@@ -400,14 +402,15 @@ class SplitCopyShared:
         if copy_proto == "ftp":
             # ftp creates 1 user process per chunk, no modulation required
             split_size = ceil(file_size / max_workers)
-        elif max_workers <= 10:
-            # 1 or 2 cpu cores, 5 or 10 workers will create 20-40 pids
+        elif max_workers == 5:
+            # 1 cpu core, 5 workers will create <= 20 pids
             # no modulation required
             split_size = ceil(file_size / max_workers)
         else:
             # scp to FreeBSD 6 based junos creates 3 user processes per chunk
             # scp to FreeBSD 10+ based junos creates 2 user processes per chunk
             # +1 user process if openssh version is >= 7.4
+            pid_count = 0
             max_pids = 40
             if sshd_version >= 7.4 and bsd_version == 6.0:
                 pid_count = 4
@@ -419,9 +422,15 @@ class SplitCopyShared:
                 pid_count = 2
             elif evo:
                 pid_count = 3
+
+            if pid_count:
+                max_workers = round(max_pids / pid_count)
             else:
-                pid_count = 4
-            max_workers = round(max_pids / pid_count)
+                # sshd config defaults
+                # Maxsessions = 10, MaxStartups = 10:30:100
+                # value here should not hit these limits
+                max_workers = 5
+
             split_size = ceil(file_size / max_workers)
 
         # concurrent.futures.ThreadPoolExecutor can be a limiting factor
@@ -437,7 +446,7 @@ class SplitCopyShared:
         return split_size, executor
 
     def mkdir_remote(self):
-        """Function that creates a tmp directory on the remote host
+        """creates a tmp directory on the remote host
         :returns remote_tmpdir:
         :type string:
         """
@@ -460,7 +469,7 @@ class SplitCopyShared:
         return remote_tmpdir
 
     def storage_check_remote(self, file_size, split_size):
-        """Function that checks whether there is enough storage space on remote node
+        """checks whether there is enough storage space on remote node
         :param file_size:
         :type int:
         :param split_size:
@@ -473,14 +482,9 @@ class SplitCopyShared:
         result, stdout = self.sshshell.run(f"df -k {self.remote_dir}")
         if not result:
             self.close(err_str="failed to determine remote disk space available")
-        df_num = len(stdout.split("\n")) - 2
-        if re.match(r"^ ", stdout.split("\n")[df_num]):
-            split_num = 2
-        else:
-            split_num = 3
         try:
-            avail_blocks = stdout.split("\n")[df_num].split()[split_num].rstrip()
-        except Exception:
+            avail_blocks = re.search(r" ([0-9]+) +[0-9]+ +[0-9]+ +", stdout).group(1)
+        except AttributeError:
             err = "unable to determine available blocks on remote host"
             self.close(err_str=err)
 
@@ -505,7 +509,7 @@ class SplitCopyShared:
                 self.close(err_str=err)
 
     def storage_check_local(self, file_size):
-        """Function that checks whether there is enough storage space on local node
+        """checks whether there is enough storage space on local node
         :param file_size:
         :type int:
         :return None:
@@ -539,7 +543,7 @@ class SplitCopyShared:
 
     @contextmanager
     def change_dir(self, cleanup=lambda: True):
-        """Function that cds into temp directory.
+        """cds into temp directory.
         Upon script exit, changes back to original directory
         and calls cleanup() to delete the temp directory
         :param cleanup:
@@ -556,8 +560,7 @@ class SplitCopyShared:
 
     @contextmanager
     def tempdir(self):
-        """Function that creates a temp directory,
-        defines how to delete directory upon script exit
+        """creates a temp directory, defines how to delete directory upon script exit
         :returns None:
         """
         self.local_tmpdir = tempfile.mkdtemp()
@@ -577,52 +580,54 @@ class SplitCopyShared:
         """
         return self.local_tmpdir
 
-    def limit_check(self, copy_proto):
-        """Function that checks the remote junos/evo hosts configuration to
-        determine whether there are any ftp or ssh connection/rate limits defined.
-        If found, these configuration lines will be deactivated
-        :param copy_proto:
-        :type string:
-        :return self.command_list or None:
+    def find_configured_limits(self, config_stanzas):
+        """Function that retrieves any configuration stazas that implement
+        rate/connection limits.
+        :param config_stanzas:
         :type list:
+        :return cli_config:
+        :type string:
         """
-        logger.info("entering limit_check()")
-        config_stanzas = ["groups", "system services", "system login"]
-        limits = ["services ssh connection-limit", "services ssh rate-limit"]
-        if copy_proto == "ftp":
-            limits.append("services ftp connection-limit")
-            limits.append("services ftp rate-limit")
-        retry_options = "system login retry-options"
-
-        # check for presence of rate/connection limits
+        logger.info("entering find_configured_limits()")
         cli_config = ""
         for stanza in config_stanzas:
             result, stdout = self.sshshell.run(
                 f'cli -c "show configuration {stanza} | display set | no-more"'
             )
             cli_config += stdout
-        if cli_config:
-            conf_list = cli_config.split("\r\n")
-        else:
-            return None
+        return cli_config
+
+    def limit_check(self, copy_proto):
+        """Function that checks the remote junos/evo hosts configuration to
+        determine whether there are any ftp or ssh connection/rate limits defined.
+        If found, these configuration lines will be deactivated
+        :param copy_proto:
+        :type string:
+        :return self.command_list:
+        :type list:
+        """
+        logger.info("entering limit_check()")
+        config_stanzas = ["groups", "system services", "system login"]
+        retry_options = "system login retry-options"
+        limits = ["services ssh connection-limit", "services ssh rate-limit"]
+        if copy_proto == "ftp":
+            limits.append("services ftp connection-limit")
+            limits.append("services ftp rate-limit")
+        cli_config = self.find_configured_limits(config_stanzas)
 
         for limit in limits:
-            if re.search(rf"^set .* {limit} [0-9]", cli_config, re.MULTILINE) is None:
-                continue
-            for conf_statement in conf_list:
-                if re.match(rf"set .* {limit} [0-9]", conf_statement):
-                    conf_line = re.sub(" [0-9]+$", "", conf_statement)
-                    conf_line = re.sub(r"^set", "deactivate", conf_line)
-                    self.command_list.append(f"{conf_line};")
-        if re.search(rf"^set .* {retry_options}", cli_config, re.MULTILINE) is not None:
-            for conf_statement in conf_list:
-                if re.match(rf"set .* {retry_options}", conf_statement):
-                    conf_line = re.match(
-                        rf"(set .* {retry_options})", conf_statement
-                    ).group(1)
-                    conf_line = re.sub(r"^set", "deactivate", conf_line)
-                    if conf_line not in self.command_list:
-                        self.command_list.append(f"{conf_line};")
+            re_limit_multiline = re.compile(rf"^set .*{limit} [0-9]", re.MULTILINE)
+            conf_list_limits = re.findall(re_limit_multiline, cli_config)
+            for conf_statement in conf_list_limits:
+                conf_line = re.sub(" [0-9]+$", "", conf_statement)
+                conf_line = re.sub(r"^set", "deactivate", conf_line)
+                self.command_list.append(f"{conf_line};")
+        re_retry_multiline = re.compile(rf"^set .*{retry_options}", re.MULTILINE)
+        conf_list_retry_options = re.findall(re_retry_multiline, cli_config)
+        for conf_statement in conf_list_retry_options:
+            conf_line = re.match(rf"(set .*{retry_options})", conf_statement).group(1)
+            conf_line = re.sub(r"^set", "deactivate", conf_line)
+            self.command_list.append(f"{conf_line};")
 
         # if limits were configured, deactivate them
         if self.command_list:
@@ -647,9 +652,7 @@ class SplitCopyShared:
                     f"configuration. output was:\n{stdout}"
                 )
                 self.close(err_str=err)
-            return self.command_list
-        else:
-            return None
+        return self.command_list
 
     def limits_rollback(self):
         """Function to revert config changes made to remote host
@@ -677,7 +680,7 @@ class SplitCopyShared:
             )
 
     def remote_cleanup(self, remote_dir=None, remote_file=None, silent=False):
-        """Functiont that deletes the tmp directory on remote host
+        """Function that deletes the tmp directory on remote host
         :param remote_dir:
         :type string:
         :param remote_file:
@@ -686,12 +689,13 @@ class SplitCopyShared:
         :type: bool
         :return None:
         """
+        logger.info("entering remote_cleanup()")
         if remote_dir:
             self.remote_dir = remote_dir
         if remote_file:
             self.remote_file = remote_file
         if not silent:
-            print(f"\r{self.pad_string('deleting remote tmp directory...')}")
+            print(f"\r{pad_string('deleting remote tmp directory...')}")
         if self.remote_tmpdir is None:
             if self.get_op:
                 self.sshshell.run(f"rm -rf /var/tmp/splitcopy_{self.remote_file}.*")
